@@ -2,9 +2,10 @@
 
 # Acknowledgements:
 #  Nicolas Economou, for his ptool suite on which this tool is inspired.
+#  http://tinyurl.com/nicolaseconomou
 
 # Process enumerator
-# Copyright (c) 2009, Mario Vilas
+# Copyright (c) 2009-2010, Mario Vilas
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -31,7 +32,7 @@
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 
-__revision__ = "$Id: plist.py 237 2009-06-16 17:13:36Z qvasimodo $"
+__revision__ = "$Id: plist.py 593 2010-01-26 20:39:38Z qvasimodo $"
 
 # TODO
 # Show more info on processes, for example the user that spawned them, or if
@@ -97,11 +98,21 @@ def main(argv):
     for pid in pid_list:
         p = s.get_process(pid)
 
-        # Special cases: PIDs 0 and 4.
+        # Special process IDs.
+        # PID 0: System Idle Process. Also has a special meaning to the
+        #        toolhelp APIs (current process).
+        # PID 4: System Integrity Group. See this forum post for more info:
+        #        http://tinyurl.com/ycza8jo
+        #        (points to social.technet.microsoft.com)
+        #        Only on XP and above
+        # PID 8: System (?) only in Windows 2000 and below AFAIK.
+        #        It's probably the same as PID 4 in XP and above.
         if pid == 0:
-            fileName = "System process"
+            fileName = "[System Idle Process]"
         elif pid == 4:
-            fileName = "System"
+            fileName = "[System]"
+        elif pid == 8:
+            fileName = "[System]"
 
         # Get the process filename (or pathname).
         else:
