@@ -1,4 +1,4 @@
-# Copyright (c) 2009, Mario Vilas
+# Copyright (c) 2009-2010, Mario Vilas
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -28,13 +28,11 @@
 """
 Debugging module.
 
-@see: U{http://apps.sourceforge.net/trac/winappdbg/wiki/wiki/Debugging}
-
-@group Instrumentation: System, Process, Thread, Module
-@group Debugging: Debug, EventHandler
+@group Debugging:
+    Debug
 """
 
-__revision__ = "$Id: debug.py 237 2009-06-16 17:13:36Z qvasimodo $"
+__revision__ = "$Id: debug.py 648 2010-03-05 21:26:13Z qvasimodo $"
 
 __all__ =   [
                 # the main debugger class
@@ -69,8 +67,6 @@ class Debug (EventDispatcher, BreakpointContainer):
     """
     The main debugger class.
 
-    @see: U{http://apps.sourceforge.net/trac/winappdbg/wiki/wiki/Debugging}
-
     @group Debugging:
         attach, detach, detach_from_all, execv, execl, clear,
         get_debugee_count, get_debugee_pids,
@@ -126,14 +122,13 @@ class Debug (EventDispatcher, BreakpointContainer):
         self.system                         = System()
         self.__bKillOnExit                  = bKillOnExit
         self.__bHostileCode                 = bHostileCode
-        self.__attachedDebugees             = set()
-        self.__startedDebugees              = set()
+        self.__attachedDebugees             = set()     # set of pids
+        self.__startedDebugees              = set()     # set of pids
 
-##        self.system.request_debug_privileges(bIgnoreExceptions = True)
-        self.system.request_debug_privileges()
+        self.system.request_debug_privileges(bIgnoreExceptions = False)
 
-##    # It's not hard to create circular references,
-##    # and since we have a destructor, we end up leaking everything.
+##    # It's hard not to create circular references,
+##    # and if we have a destructor, we can end up leaking everything.
 ##    # It's best to code the debugging loop properly to always
 ##    # stop the debugger before going out of scope.
 ##    def __del__(self):
@@ -144,6 +139,21 @@ class Debug (EventDispatcher, BreakpointContainer):
 ####            traceback.print_exc()
 ####            print
 
+    def __enter__(self):
+        """
+        Compatibility with the "C{with}" Python statement.
+        """
+        return self
+
+    def __exit__(self, type, value, traceback):
+        """
+        Compatibility with the "C{with}" Python statement.
+        """
+        try:
+            self.stop()
+        except Exception, e:
+            pass
+
     def __len__(self):
         """
         @rtype:  int
@@ -184,7 +194,7 @@ class Debug (EventDispatcher, BreakpointContainer):
         else:
             aProcess = self.system.get_process(dwProcessId)
 
-        # XXX
+        # XXX HACK
         # Scan the process threads and loaded modules.
         # This is prefered because the thread and library events do not
         # properly give some information, like the filename for each module.
@@ -193,27 +203,27 @@ class Debug (EventDispatcher, BreakpointContainer):
 
         return aProcess
 
-    def detach(self, dwProcessId, bIgnoreExceptions = False):
+    def __cleanup_process(self, dwProcessId, bIgnoreExceptions = False):
         """
-        Detaches from a process currently being debugged.
+        Perform the necessary cleanup of a process about to be killed or
+        detached from.
 
-        @see: L{attach}, L{detach_from_all}
+        This private method is called by L{kill} and L{detach}.
 
         @type  dwProcessId: int
-        @param dwProcessId: Global ID of a process to detach from.
+        @param dwProcessId: Global ID of a process to kill.
 
         @type  bIgnoreExceptions: bool
         @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
-            raised when detaching.
+            raised when killing the process.
 
         @raise WindowsError: Raises an exception on error, unless
             C{bIgnoreExceptions} is C{True}.
         """
 
-        # Disable all breakpoints in the process.
-        # XXX maybe they should be erased?
+        # Erase all breakpoints in the process.
         try:
-            self.disable_process_breakpoints(dwProcessId)
+            self.erase_process_breakpoints(dwProcessId)
         except Exception:
             if not bIgnoreExceptions:
                 raise
@@ -241,15 +251,110 @@ class Debug (EventDispatcher, BreakpointContainer):
 ##            traceback.print_exc()
 ##            print
 
-        # Detach from the process.
+        # Clear and remove the process from the snapshot.
+        # If the user wants to do something with it after detaching
+        # a new Process instance must be created.
         try:
-            win32.DebugActiveProcessStop(dwProcessId)
+            self.system.get_process(dwProcessId).clear()
+        except Exception:
+            if not bIgnoreExceptions:
+                raise
+        try:
+            # XXX HACK
+            self.system._ProcessContainer__del_process(dwProcessId)
+        except Exception:
+            if not bIgnoreExceptions:
+                raise
+
+    def kill(self, dwProcessId, bIgnoreExceptions = False):
+        """
+        Kills a process currently being debugged.
+
+        @see: L{detach}
+
+        @type  dwProcessId: int
+        @param dwProcessId: Global ID of a process to kill.
+
+        @type  bIgnoreExceptions: bool
+        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
+            raised when killing the process.
+
+        @raise WindowsError: Raises an exception on error, unless
+            C{bIgnoreExceptions} is C{True}.
+        """
+
+        # XXX FIXME
+        # what happens if we didn't know the process ???
+        # more validation is needed here!
+
+        # Keep a reference to the process. We'll need it later.
+        aProcess = self.system.get_process(dwProcessId)
+
+        # Cleanup all data referring to the process.
+        self.__cleanup_process(dwProcessId)
+
+        # Kill the process.
+        try:
+            aProcess.kill()
         except Exception:
              if not bIgnoreExceptions:
                 raise
 ##            traceback.print_exc()
 ##            print
 
+        # Cleanup what remains of the process data.
+        aProcess.clear()
+
+    def detach(self, dwProcessId, bIgnoreExceptions = False):
+        """
+        Detaches from a process currently being debugged.
+
+        @note: On Windows 2000 and below the process is killed.
+
+        @see: L{attach}, L{detach_from_all}
+
+        @type  dwProcessId: int
+        @param dwProcessId: Global ID of a process to detach from.
+
+        @type  bIgnoreExceptions: bool
+        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
+            raised when detaching.
+
+        @raise WindowsError: Raises an exception on error, unless
+            C{bIgnoreExceptions} is C{True}.
+        """
+
+        # XXX FIXME
+        # what happens if we didn't know the process ???
+        # more validation is needed here!
+
+        # Keep a reference to the process. We'll need it later.
+        aProcess = self.system.get_process(dwProcessId)
+
+        # Cleanup all data referring to the process.
+        self.__cleanup_process(dwProcessId)
+
+        # Detach from the process.
+        # On Windows 2000 and before, kill the process.
+        try:
+            win32.DebugActiveProcessStop(dwProcessId)
+        except AttributeError:
+            try:
+                aProcess.kill()
+            except Exception:
+                 if not bIgnoreExceptions:
+                    raise
+    ##            traceback.print_exc()
+    ##            print
+        except Exception:
+             if not bIgnoreExceptions:
+                raise
+##            traceback.print_exc()
+##            print
+
+        # Cleanup what remains of the process data.
+        aProcess.clear()
+
     def detach_from_all(self, bIgnoreExceptions = False):
         """
         Detaches from all processes currently being debugged.
@@ -268,7 +373,9 @@ class Debug (EventDispatcher, BreakpointContainer):
 
     def execv(self, argv,                                    bConsole = False,
                                                               bFollow = False,
-                                                           bSuspended = False):
+                                                           bSuspended = False,
+                                                      bInheritHandles = False,
+                                                    dwParentProcessId = None):
         """
         Starts a new process for debugging.
 
@@ -283,13 +390,24 @@ class Debug (EventDispatcher, BreakpointContainer):
 
         @type  bConsole: bool
         @param bConsole: True to inherit the console of the debugger.
+            Defaults to C{False}.
 
         @type  bFollow: bool
         @param bFollow: C{True} to automatically attach to child processes.
+            Defaults to C{False}.
 
         @type  bSuspended: bool
         @param bSuspended: C{True} to suspend the main thread before any code
-            is executed in the debugee.
+            is executed in the debugee. Defaults to C{False}.
+
+        @type  bInheritHandles: bool
+        @param bInheritHandles: C{True} if the new process should inherit it's
+            parent process' handles. Defaults to C{False}.
+
+        @type  dwParentProcessId: int or None
+        @param dwParentProcessId: C{None} if the debugger process should be the
+            parent process (default), or a process ID to forcefully set as the
+            debugee's parent (only available for Windows Vista and above).
 
         @rtype:  L{Process}
         @return: A new Process object.
@@ -299,11 +417,15 @@ class Debug (EventDispatcher, BreakpointContainer):
         lpCmdLine = self.system.argv_to_cmdline(argv)
         return self.execl(lpCmdLine,   bConsole = bConsole,
                                         bFollow = bFollow,
-                                     bSuspended = bSuspended)
+                                     bSuspended = bSuspended,
+                                bInheritHandles = bInheritHandles,
+                              dwParentProcessId = dwParentProcessId)
 
     def execl(self, lpCmdLine,                               bConsole = False,
                                                               bFollow = False,
-                                                           bSuspended = False):
+                                                           bSuspended = False,
+                                                      bInheritHandles = False,
+                                                    dwParentProcessId = None):
         """
         Starts a new process for debugging.
 
@@ -321,13 +443,24 @@ class Debug (EventDispatcher, BreakpointContainer):
 
         @type  bConsole: bool
         @param bConsole: C{True} to inherit the console of the debugger.
+            Defaults to C{False}.
 
         @type  bFollow: bool
         @param bFollow: C{True} to automatically attach to child processes.
+            Defaults to C{False}.
 
         @type  bSuspended: bool
         @param bSuspended: C{True} to suspend the main thread before any code
-            is executed in the debugee.
+            is executed in the debugee. Defaults to C{False}.
+
+        @type  bInheritHandles: bool
+        @param bInheritHandles: C{True} if the new process should inherit it's
+            parent process' handles. Defaults to C{False}.
+
+        @type  dwParentProcessId: int or None
+        @param dwParentProcessId: C{None} if the debugger process should be the
+            parent process (default), or a process ID to forcefully set as the
+            debugee's parent (only available for Windows Vista and above).
 
         @rtype:  L{Process}
         @return: A new Process object.
@@ -335,10 +468,12 @@ class Debug (EventDispatcher, BreakpointContainer):
         @raise WindowsError: Raises an exception on error.
         """
         aProcess = self.system.start_process(lpCmdLine,
-            bConsole    = bConsole,
-            bDebug      = True,
-            bFollow     = bFollow,
-            bSuspended  = bSuspended
+            bConsole          = bConsole,
+            bDebug            = True,
+            bFollow           = bFollow,
+            bSuspended        = bSuspended,
+            bInheritHandles   = bInheritHandles,
+            dwParentProcessId = dwParentProcessId,
         )
 
         self.__startedDebugees.add( aProcess.get_pid() )
@@ -432,6 +567,26 @@ class Debug (EventDispatcher, BreakpointContainer):
         except WindowsError:
             pass
 
+##        # XXX Just for testing, ignore this...
+##        print "ContinueDebugEvent(%d, %d, %s)" % (dwProcessId, dwThreadId, {
+##            win32.DBG_CONTINUE                    : "DBG_CONTINUE",
+##            win32.DBG_EXCEPTION_HANDLED           : "DBG_EXCEPTION_HANDLED",
+##            win32.DBG_EXCEPTION_NOT_HANDLED       : "DBG_EXCEPTION_NOT_HANDLED",
+##            win32.DBG_TERMINATE_THREAD            : "DBG_TERMINATE_THREAD",
+##            win32.DBG_TERMINATE_PROCESS           : "DBG_TERMINATE_PROCESS",
+##        }.get(dwContinueStatus, hex(dwContinueStatus)))
+
+##        # XXX Just for testing, ignore this...
+##        if hasattr(event, 'is_noncontinuable') and event.is_noncontinuable():
+##            from textio import DebugLog
+##            print DebugLog.log_event(event, "Warning: noncontinuable event!")
+
+        # XXX TODO
+        # * try to execute the UnhandledExceptionFilter for second chance
+        # exceptions, at least when in hostile mode (in normal mode it would be
+        # breaking compatibility, as users may actually expect second chance
+        # exceptions to be raised again).
+
         # Continue execution of the debugee.
         win32.ContinueDebugEvent(dwProcessId, dwThreadId, dwContinueStatus)
 
@@ -455,30 +610,39 @@ class Debug (EventDispatcher, BreakpointContainer):
         @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
             raised when detaching.
         """
-        # All these try / except / finally blocks may be masking some errors,
-        # but this way we get a better cleanup. Like that battery-powered
-        # rabbit, it just keeps going, and going, and going.
+        # I wish I knew a more pythonic way of doing this :(
+        has_event = False
         try:
-            if self.__bKillOnExit:
-                for pid in self.get_debugee_pids():
-                    try:
-                        self.system.get_process(pid).kill()
-                    except Exception:
-                        if not bIgnoreExceptions:
-                            raise
+            has_event = bool(event)
+        except Exception:
+            if not bIgnoreExceptions:
+                raise
+        if has_event:
             try:
-                if event:
-                    try:
-                        try:
-                            pid = event.get_pid()
-                            self.disable_process_breakpoints(pid)
-                        finally:
-                            self.cont(event)
-                    except Exception:
-                        if not bIgnoreExceptions:
-                            raise
-            finally:
-                self.detach_from_all(bIgnoreExceptions)
+                pid = event.get_pid()
+                self.disable_process_breakpoints(pid)
+            except Exception:
+                if not bIgnoreExceptions:
+                    raise
+            try:
+                tid = event.get_tid()
+                self.disable_thread_breakpoints(tid)
+            except Exception:
+                if not bIgnoreExceptions:
+                    raise
+            try:
+                event.continueDebugEvent = win32.DBG_CONTINUE
+                self.cont(event)
+            except Exception:
+                if not bIgnoreExceptions:
+                    raise
+        try:
+            self.detach_from_all(bIgnoreExceptions)
+        except Exception:
+            if not bIgnoreExceptions:
+                raise
+        try:
+            self.system.clear()
         except Exception:
             if not bIgnoreExceptions:
                 raise
@@ -516,7 +680,7 @@ class Debug (EventDispatcher, BreakpointContainer):
         Simple debugging loop.
 
         This debugging loop is meant to be useful for most simple scripts.
-        It iterates as long as there is at least one debuguee, or an exception
+        It iterates as long as there is at least one debugee, or an exception
         is raised. Multiple calls are allowed.
 
         This is a trivial example script::
@@ -824,7 +988,7 @@ class Debug (EventDispatcher, BreakpointContainer):
         @return: C{True} to call the user-defined handle, C{False} otherwise.
         """
         if event.is_first_chance():
-            event.continueStatus = win32.DBG_CONTINUE
+            event.continueStatus = win32.DBG_EXCEPTION_HANDLED
         return True
 
     def notify_ms_vc_exception(self, event):
@@ -847,7 +1011,7 @@ class Debug (EventDispatcher, BreakpointContainer):
         dwType = event.get_exception_information(0)
         if dwType == 0x1000:
             pszName     = event.get_exception_information(1)
-            dwThreadID  = event.get_exception_information(2)
+            dwThreadId  = event.get_exception_information(2)
             dwFlags     = event.get_exception_information(3)
 
             aProcess = event.get_process()
@@ -861,7 +1025,7 @@ class Debug (EventDispatcher, BreakpointContainer):
                     aThread = aProcess.get_thread(dwThreadId)
                 else:
                     aThread = Thread(dwThreadId)
-                    aProcess._Process__add_thread(aThread)
+                    aProcess._ThreadContainer__add_thread(aThread)
 
 ##                if aThread.get_name() is None:
 ##                    aThread.set_name(szName)
