@@ -1,4 +1,4 @@
-# Copyright (c) 2009, Mario Vilas
+# Copyright (c) 2009-2010, Mario Vilas
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,12 @@ Event handling module.
 
 @see: U{http://apps.sourceforge.net/trac/winappdbg/wiki/Debugging}
 
-@group Event objects:
+@group Debugging:
+    EventHandler
+
+@group Debug events:
+    EventFactory,
+    EventDispatcher,
     Event,
     NoEvent,
     CreateProcessEvent,
@@ -44,19 +49,36 @@ Event handling module.
     ExceptionEvent
 """
 
-__revision__ = "$Id: event.py 237 2009-06-16 17:13:36Z qvasimodo $"
+__revision__ = "$Id: event.py 689 2010-06-20 13:56:12Z qvasimodo $"
 
 __all__ = [
             # Factory of Event objects and all of it's subclasses.
             # Users should not need to instance Event objects directly.
             'EventFactory',
 
+            # Event dispatcher used internally by the Debug class.
+            'EventDispatcher',
+
             # Base class for user-defined event handlers.
             'EventHandler',
 
             # Dummy event object that can be used as a placeholder.
             # It's never returned by the EventFactory.
             'NoEvent',
+
+            # Base class for event objects.
+            'Event',
+
+            # Event objects.
+            'CreateProcessEvent',
+            'CreateThreadEvent',
+            'ExitProcessEvent',
+            'ExitThreadEvent',
+            'LoadDLLEvent',
+            'UnloadDLLEvent',
+            'OutputDebugStringEvent',
+            'RIPEvent',
+            'ExceptionEvent'
           ]
 
 import win32
@@ -66,6 +88,7 @@ from system import Module, Thread, Process, PathOperations
 from textio import HexDump
 
 import ctypes
+##import weakref
 
 #==============================================================================
 
@@ -76,6 +99,7 @@ class Event (object):
     @type eventMethod: str
     @cvar eventMethod:
         Method name to call when using L{EventHandler} subclasses.
+        Used internally.
 
     @type eventName: str
     @cvar eventName:
@@ -110,10 +134,20 @@ class Event (object):
         @type  raw: L{DEBUG_EVENT}
         @param raw: Raw DEBUG_EVENT structure as used by the Win32 API.
         """
+##        self.__debug        = weakref.ref(debug)
         self.debug          = debug
         self.raw            = raw
         self.continueStatus = win32.DBG_EXCEPTION_NOT_HANDLED
 
+##    @property
+##    def debug(self):
+##        """
+##        @rtype  debug: L{Debug}
+##        @return debug:
+##            Debug object that received the event.
+##        """
+##        return self.__debug()
+
     def get_event_name(self):
         """
         @rtype:  str
@@ -135,18 +169,18 @@ class Event (object):
         """
         return self.raw.dwDebugEventCode
 
-    # Compatibility with version 1.0
-    # XXX to be removed in version 1.3
-    def get_code(self):
-        """
-        Alias of L{get_event_code} for backwards compatibility
-        with WinAppDbg version 1.0.
-        Will be phased out in the next version.
-
-        @rtype:  int
-        @return: Debug event code as defined in the Win32 API.
-        """
-        return self.get_event_code()
+##    # Compatibility with version 1.0
+##    # XXX to be removed in version 1.4
+##    def get_code(self):
+##        """
+##        Alias of L{get_event_code} for backwards compatibility
+##        with WinAppDbg version 1.0.
+##        Will be phased out in the next version.
+##
+##        @rtype:  int
+##        @return: Debug event code as defined in the Win32 API.
+##        """
+##        return self.get_event_code()
 
     def get_pid(self):
         """
@@ -173,15 +207,13 @@ class Event (object):
         @rtype:  L{Process}
         @return: Process where the event occured.
         """
-        # We can't assume the Process object will be in the System snapshot.
-        # The user may have cleared or otherwise modified the snapshot.
-        # Also some process creation events are missed on Wine.
         pid     = self.get_pid()
         system  = self.debug.system
         if system.has_process(pid):
             process = system.get_process(pid)
         else:
-##            print "Process notification missed! ID: %x" % pid       # XXX
+            # XXX HACK
+            # The process object was missing for some reason, so make a new one.
             process = Process(pid)
             system._ProcessContainer__add_process(process)
 ##            process.scan_threads()    # not needed
@@ -195,14 +227,13 @@ class Event (object):
         @rtype:  L{Thread}
         @return: Thread where the event occured.
         """
-        # We can't assume the Thread object will be in the Process snapshot.
-        # See the comments of get_process.
         tid     = self.get_tid()
         process = self.get_process()
         if process.has_thread(tid):
             thread = process.get_thread(tid)
         else:
-##            print "Thread notification missed! ID: %x" % tid        # XXX
+            # XXX HACK
+            # The thread object was missing for some reason, so make a new one.
             thread = Thread(tid)
             process._ThreadContainer__add_thread(thread)
         return thread
@@ -264,9 +295,16 @@ class ExceptionEvent (Event):
     @type breakpoint: L{Breakpoint}
     @ivar breakpoint:
         If the exception was caused by one of our breakpoints, this member
-        contains the a reference to the breakpoint object. Otherwise it's
-        not defined. It should only be used from the condition or action
-        callback routines, instead of the event handler.
+        contains a reference to the breakpoint object. Otherwise it's not
+        defined. It should only be used from the condition or action callback
+        routines, instead of the event handler.
+
+    @type hook: L{Hook}
+    @ivar hook:
+        If the exception was caused by a function hook, this member contains a
+        reference to the hook object. Otherwise it's not defined. It should
+        only be used from the hook callback routines, instead of the event
+        handler.
     """
 
     eventName        = 'Exception event'
@@ -296,6 +334,7 @@ class ExceptionEvent (Event):
         win32.EXCEPTION_GUARD_PAGE                : 'guard_page',
         win32.EXCEPTION_INVALID_HANDLE            : 'invalid_handle',
         win32.EXCEPTION_POSSIBLE_DEADLOCK         : 'possible_deadlock',
+        win32.EXCEPTION_WX86_BREAKPOINT           : 'wow64_breakpoint',
         win32.CONTROL_C_EXIT                      : 'control_c_exit',
         win32.DBG_CONTROL_C                       : 'debug_control_c',
         win32.MS_VC_EXCEPTION                     : 'ms_vc_exception',
@@ -325,6 +364,7 @@ class ExceptionEvent (Event):
         win32.EXCEPTION_GUARD_PAGE                : 'EXCEPTION_GUARD_PAGE',
         win32.EXCEPTION_INVALID_HANDLE            : 'EXCEPTION_INVALID_HANDLE',
         win32.EXCEPTION_POSSIBLE_DEADLOCK         : 'EXCEPTION_POSSIBLE_DEADLOCK',
+        win32.EXCEPTION_WX86_BREAKPOINT           : 'EXCEPTION_WX86_BREAKPOINT',
         win32.CONTROL_C_EXIT                      : 'CONTROL_C_EXIT',
         win32.DBG_CONTROL_C                       : 'DBG_CONTROL_C',
         win32.MS_VC_EXCEPTION                     : 'MS_VC_EXCEPTION',
@@ -333,7 +373,7 @@ class ExceptionEvent (Event):
     __exceptionDescription = {
         win32.EXCEPTION_ACCESS_VIOLATION          : 'Access violation',
         win32.EXCEPTION_ARRAY_BOUNDS_EXCEEDED     : 'Array bounds exceeded',
-        win32.EXCEPTION_BREAKPOINT                : 'Breakpoint event',
+        win32.EXCEPTION_BREAKPOINT                : 'Breakpoint',
         win32.EXCEPTION_DATATYPE_MISALIGNMENT     : 'Datatype misalignment',
         win32.EXCEPTION_FLT_DENORMAL_OPERAND      : 'Float denormal operand',
         win32.EXCEPTION_FLT_DIVIDE_BY_ZERO        : 'Float divide by zero',
@@ -354,6 +394,7 @@ class ExceptionEvent (Event):
         win32.EXCEPTION_GUARD_PAGE                : 'Guard page hit',
         win32.EXCEPTION_INVALID_HANDLE            : 'Invalid handle',
         win32.EXCEPTION_POSSIBLE_DEADLOCK         : 'Possible deadlock',
+        win32.EXCEPTION_WX86_BREAKPOINT           : 'WOW64 breakpoint',
         win32.CONTROL_C_EXIT                      : 'Control-C exit',
         win32.DBG_CONTROL_C                       : 'Debug Control-C',
         win32.MS_VC_EXCEPTION                     : 'Microsoft Visual C exception',
@@ -428,7 +469,10 @@ class ExceptionEvent (Event):
         @rtype:  int
         @return: Memory address where the exception occured.
         """
-        return self.raw.u.Exception.ExceptionRecord.ExceptionAddress
+        address = self.raw.u.Exception.ExceptionRecord.ExceptionAddress
+        if address is None:
+            address = 0
+        return address
 
     def get_exception_information(self, index):
         """
@@ -441,7 +485,10 @@ class ExceptionEvent (Event):
         if index < 0 or index > win32.EXCEPTION_MAXIMUM_PARAMETERS:
             raise IndexError, "Array index out of range: %s" % repr(index)
         info = self.raw.u.Exception.ExceptionRecord.ExceptionInformation
-        return info[index]
+        value = info[index]
+        if value is None:
+            value = 0
+        return value
 
     def get_exception_information_as_list(self):
         """
@@ -449,15 +496,77 @@ class ExceptionEvent (Event):
         @return: Exception information block.
         """
         info = self.raw.u.Exception.ExceptionRecord.ExceptionInformation
-        return [ info[i] for i in xrange(0, win32.EXCEPTION_MAXIMUM_PARAMETERS) ]
+        data = list()
+        for index in xrange(0, win32.EXCEPTION_MAXIMUM_PARAMETERS):
+            value = info[index]
+            if value is None:
+                value = 0
+            data.append(value)
+        return data
 
-    def get_access_violation_type(self):
-        if self.get_exception_code() != win32.EXCEPTION_ACCESS_VIOLATION:
-            msg = ("This method is only meaningful"
-                   " for access violation exceptions.")
-            raise Exception, msg
+    def get_fault_type(self):
+        """
+        @rtype:  int
+        @return: Access violation type.
+            Should be one of the following constants:
+
+             - L{win32.EXCEPTION_READ_FAULT}
+             - L{win32.EXCEPTION_WRITE_FAULT}
+             - L{win32.EXCEPTION_EXECUTE_FAULT}
+
+        @note: This method is only meaningful for access violation exceptions,
+            in-page memory error exceptions and guard page exceptions.
+
+        @raise NotImplementedError: Wrong kind of exception.
+        """
+        if self.get_exception_code() not in (win32.EXCEPTION_ACCESS_VIOLATION,
+                    win32.EXCEPTION_IN_PAGE_ERROR, win32.EXCEPTION_GUARD_PAGE):
+            msg = "This method is not meaningful for %s."
+            raise NotImplementedError, msg % self.get_exception_name()
         return self.get_exception_information(0)
 
+    def get_fault_address(self):
+        """
+        @rtype:  int
+        @return: Access violation memory address.
+
+        @note: This method is only meaningful for access violation exceptions,
+            in-page memory error exceptions and guard page exceptions.
+
+        @raise NotImplementedError: Wrong kind of exception.
+        """
+        if self.get_exception_code() not in (win32.EXCEPTION_ACCESS_VIOLATION,
+                    win32.EXCEPTION_IN_PAGE_ERROR, win32.EXCEPTION_GUARD_PAGE):
+            msg = "This method is not meaningful for %s."
+            raise NotImplementedError, msg % self.get_exception_name()
+        return self.get_exception_information(1)
+
+    def get_ntstatus_code(self):
+        """
+        @rtype:  int
+        @return: NTSTATUS status code that caused the exception.
+
+        @note: This method is only meaningful for in-page memory error
+            exceptions.
+
+        @raise NotImplementedError: Not an in-page memory error.
+        """
+        if self.get_exception_code() != win32.EXCEPTION_IN_PAGE_ERROR:
+            msg = "This method is only meaningful "\
+                  "for in-page memory error exceptions."
+            raise NotImplementedError, msg
+        return self.get_exception_information(2)
+
+    def is_nested(self):
+        """
+        @rtype:  bool
+        @return: Returns C{True} if there are additional exception records
+            associated with this exception. This would mean the exception
+            is nested, that is, it was triggered while trying to handle
+            at least one previous exception.
+        """
+        return bool(self.raw.u.Exception.ExceptionRecord.ExceptionRecord)
+
     def get_raw_exception_record_list(self):
         """
         Traverses the exception record linked list and builds a Python list.
@@ -466,7 +575,7 @@ class ExceptionEvent (Event):
         happens when an exception is raised in the debugee while trying to
         handle a previous exception.
 
-        @rtype:  list( L{EXCEPTION_RECORD} )
+        @rtype:  list( L{win32.EXCEPTION_RECORD} )
         @return:
             List of raw exception record structures as used by the Win32 API.
 
@@ -474,28 +583,57 @@ class ExceptionEvent (Event):
             never empty. All other methods of this class read from the first
             exception record only, that is, the most recent exception.
         """
-
         # The first EXCEPTION_RECORD is contained in EXCEPTION_DEBUG_INFO.
-        record = self.raw.u.Exception.ExceptionRecord
-        nested = [ record ]
-
         # The remaining EXCEPTION_RECORD structures are linked by pointers.
-        while record.ExceptionRecord:
-            record = record.ExceptionRecord.contents
+        nested = list()
+        record = self.raw.u.Exception
+        while True:
+            record = record.ExceptionRecord
+            if not record:
+                break
             nested.append(record)
-
-        # Return the list of nested exceptions.
         return nested
 
-    # TODO
-    # Return the nested exceptions as a list of ExceptionEvent objects.
-    # The first element is always "self".
-    # Each element contains a pointer to the next exception record.
-    # New raw structures may have to be allocated for each object,
-    # but it's OK to reuse pointers since they're supposed to be read only.
-    # (Perhaps a custom DEBUG_EVENT has to be defined, where the first record
-    # is a pointer instead of an embedded structure).
-##    def get_nested_exceptions(self):
+    def get_nested_exceptions(self):
+        """
+        Traverses the exception record linked list and builds a Python list.
+
+        Nested exception records are received for nested exceptions. This
+        happens when an exception is raised in the debugee while trying to
+        handle a previous exception.
+
+        @rtype:  list( L{ExceptionEvent} )
+        @return:
+            List of ExceptionEvent objects representing each exception record
+            found in this event.
+
+            There is always at least one exception record, so the list is
+            never empty. All other methods of this class read from the first
+            exception record only, that is, the most recent exception.
+        """
+        # The list always begins with ourselves.
+        # Just put a reference to "self" as the first element,
+        # and start looping from the second exception record.
+        nested = [ self ]
+        raw = self.raw
+        dwDebugEventCode = raw.dwDebugEventCode
+        dwProcessId      = raw.dwProcessId
+        dwThreadId       = raw.dwThreadId
+        dwFirstChance    = raw.u.Exception.dwFirstChance
+        record           = raw.u.Exception.ExceptionRecord
+        while True:
+            record = record.ExceptionRecord
+            if not record:
+                break
+            raw = win32.DEBUG_EVENT()
+            raw.dwDebugEventCode            = dwDebugEventCode
+            raw.dwProcessId                 = dwProcessId
+            raw.dwThreadId                  = dwThreadId
+            raw.u.Exception.ExceptionRecord = record
+            raw.u.Exception.dwFirstChance   = dwFirstChance
+            event = EventFactory.get(self.debug, raw)
+            nested.append(event)
+        return nested
 
 #==============================================================================
 
@@ -512,16 +650,14 @@ class CreateThreadEvent (Event):
         """
         @rtype:  L{ThreadHandle}
         @return: Thread handle received from the system.
-            If it's a valid handle, a new L{ThreadHandle} object is created.
-            Otherwise, the method returns C{INVALID_HANDLE_VALUE}.
-        @note: This method never returns C{NULL}.
+            Returns C{None} if the handle is not available.
         """
         # The handle doesn't need to be closed.
         # See http://msdn.microsoft.com/en-us/library/ms681423(VS.85).aspx
         hThread = self.raw.u.CreateThread.hThread
-        if hThread == win32.NULL:
-            hThread = win32.INVALID_HANDLE_VALUE
-        elif hThread != win32.INVALID_HANDLE_VALUE:
+        if hThread in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+            hThread = None
+        else:
             hThread = ThreadHandle(hThread, False)
         return hThread
 
@@ -557,22 +693,20 @@ class CreateProcessEvent (Event):
 
     def get_file_handle(self):
         """
-        @rtype:  L{FileHandle}
-        @return: File handle to the main module.
-            If it's a valid handle, a new L{FileHandle} object is created.
-            Otherwise, the method returns C{INVALID_HANDLE_VALUE}.
-        @note: This method never returns C{NULL}.
+        @rtype:  L{FileHandle} or None
+        @return: File handle to the main module, received from the system.
+            Returns C{None} if the handle is not available.
         """
         # This handle DOES need to be closed.
         # Therefore we must cache it so it doesn't
         # get closed after the first call.
-        if hasattr(self, '_CreateProcessEvent__hFile'):
+        try:
             hFile = self.__hFile
-        else:
+        except AttributeError:
             hFile = self.raw.u.CreateProcessInfo.hFile
-            if hFile == win32.NULL:
-                hFile = win32.INVALID_HANDLE_VALUE
-            elif hFile != win32.INVALID_HANDLE_VALUE:
+            if hFile in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+                hFile = None
+            else:
                 hFile = FileHandle(hFile, True)
             self.__hFile = hFile
         return hFile
@@ -581,16 +715,14 @@ class CreateProcessEvent (Event):
         """
         @rtype:  L{ProcessHandle}
         @return: Process handle received from the system.
-            If it's a valid handle, a new L{ProcessHandle} object is created.
-            Otherwise, the method returns C{INVALID_HANDLE_VALUE}.
-        @note: This method never returns C{NULL}.
+            Returns C{None} if the handle is not available.
         """
         # The handle doesn't need to be closed.
         # See http://msdn.microsoft.com/en-us/library/ms681423(VS.85).aspx
         hProcess = self.raw.u.CreateProcessInfo.hProcess
-        if hProcess == win32.NULL:
-            hProcess = win32.INVALID_HANDLE_VALUE
-        elif hProcess != win32.INVALID_HANDLE_VALUE:
+        if hProcess in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+            hProcess = None
+        else:
             hProcess = ProcessHandle(hProcess, False)
         return hProcess
 
@@ -598,16 +730,14 @@ class CreateProcessEvent (Event):
         """
         @rtype:  L{ThreadHandle}
         @return: Thread handle received from the system.
-            If it's a valid handle, a new L{ThreadHandle} object is created.
-            Otherwise, the method returns C{INVALID_HANDLE_VALUE}.
-        @note: This method never returns C{NULL}.
+            Returns C{None} if the handle is not available.
         """
         # The handle doesn't need to be closed.
         # See http://msdn.microsoft.com/en-us/library/ms681423(VS.85).aspx
         hThread = self.raw.u.CreateProcessInfo.hThread
-        if hThread == win32.NULL:
-            hThread = win32.INVALID_HANDLE_VALUE
-        elif hThread != win32.INVALID_HANDLE_VALUE:
+        if hThread in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+            hThread = None
+        else:
             hThread = ThreadHandle(hThread, False)
         return hThread
 
@@ -658,7 +788,10 @@ class CreateProcessEvent (Event):
         """
 
         # Try to get the filename from the file handle.
-        szFilename = self.get_file_handle().get_filename()
+        szFilename = None
+        hFile = self.get_file_handle()
+        if hFile:
+            szFilename = hFile.get_filename()
         if not szFilename:
 
             # Try to get it from CREATE_PROCESS_DEBUG_INFO.lpImageName
@@ -671,6 +804,10 @@ class CreateProcessEvent (Event):
                 fUnicode    = bool( self.raw.u.CreateProcessInfo.fUnicode )
                 szFilename  = aProcess.peek_string(lpFilename, fUnicode)
 
+                # XXX TODO
+                # Sometimes the filename is relative (ntdll.dll, kernel32.dll).
+                # It could be converted to an absolute pathname (SearchPath).
+
             # Try to get it from Process.get_image_name().
             if not szFilename:
                 szFilename = aProcess.get_image_name()
@@ -780,26 +917,36 @@ class LoadDLLEvent (Event):
         @rtype:  L{Module}
         @return: Module object for the newly loaded DLL.
         """
-        return self.get_process().get_module( self.get_module_base() )
+        lpBaseOfDll = self.get_module_base()
+        aProcess    = self.get_process()
+        if aProcess.has_module(lpBaseOfDll):
+            aModule = aProcess.get_module(lpBaseOfDll)
+        else:
+            # XXX HACK
+            # For some reason the module object is missing, so make a new one.
+            aModule = Module(lpBaseOfDll,
+                             hFile    = self.get_file_handle(),
+                             fileName = get_filename(),
+                             process  = aProcess)
+            aProcess.__ModuleContainer_add_module(aModule)
+        return aModule
 
     def get_file_handle(self):
         """
-        @rtype:  L{FileHandle}
-        @return: File handle to the newly loaded DLL.
-            If it's a valid handle, a new L{FileHandle} object is created.
-            Otherwise, the method returns C{INVALID_HANDLE_VALUE}.
-        @note: This method never returns C{NULL}.
+        @rtype:  L{FileHandle} or None
+        @return: File handle to the newly loaded DLL received from the system.
+            Returns C{None} if the handle is not available.
         """
         # This handle DOES need to be closed.
         # Therefore we must cache it so it doesn't
         # get closed after the first call.
-        if hasattr(self, '_CreateProcessEvent__hFile'):
+        try:
             hFile = self.__hFile
-        else:
+        except AttributeError:
             hFile = self.raw.u.LoadDll.hFile
-            if hFile == win32.NULL:
-                hFile = win32.INVALID_HANDLE_VALUE
-            elif hFile != win32.INVALID_HANDLE_VALUE:
+            if hFile in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+                hFile = None
+            else:
                 hFile = FileHandle(hFile, True)
             self.__hFile = hFile
         return hFile
@@ -811,22 +958,25 @@ class LoadDLLEvent (Event):
         the newly loaded module. However, sometimes that's not
         possible, and C{None} is returned instead.
         """
+        szFilename = None
+
+        # Try to get it from LOAD_DLL_DEBUG_INFO.lpImageName
+        # It's NULL or *NULL most of the times, see MSDN:
+        # http://msdn.microsoft.com/en-us/library/ms679286(VS.85).aspx
+        aProcess = self.get_process()
+        lpRemoteFilenamePtr = self.raw.u.LoadDll.lpImageName
+        if lpRemoteFilenamePtr:
+            lpFilename  = aProcess.peek_uint(lpRemoteFilenamePtr)
+            fUnicode    = bool( self.raw.u.LoadDll.fUnicode )
+            szFilename  = aProcess.peek_string(lpFilename, fUnicode)
+            if not szFilename:
+                szFilename = None
 
         # Try to get the filename from the file handle.
-        szFilename = self.get_file_handle().get_filename()
         if not szFilename:
-
-            # Try to get it from LOAD_DLL_DEBUG_INFO.lpImageName
-            # It's NULL or *NULL most of the times, see MSDN:
-            # http://msdn.microsoft.com/en-us/library/ms679286(VS.85).aspx
-            aProcess = self.get_process()
-            lpRemoteFilenamePtr = self.raw.u.LoadDll.lpImageName
-            if lpRemoteFilenamePtr:
-                lpFilename  = aProcess.peek_uint(lpRemoteFilenamePtr)
-                fUnicode    = bool( self.raw.u.LoadDll.fUnicode )
-                szFilename  = aProcess.peek_string(lpFilename, fUnicode)
-                if not szFilename:
-                    szFilename = None
+            hFile = self.get_file_handle()
+            if hFile:
+                szFilename = hFile.get_filename()
 
         # Return the filename, or None on error.
         return szFilename
@@ -854,7 +1004,25 @@ class UnloadDLLEvent (Event):
         @rtype:  L{Module}
         @return: Module object for the recently unloaded DLL.
         """
-        return self.get_process().get_module( self.get_module_base() )
+        lpBaseOfDll = self.get_module_base()
+        aProcess    = self.get_process()
+        if aProcess.has_module(lpBaseOfDll):
+            aModule = aProcess.get_module(lpBaseOfDll)
+        else:
+            aModule = Module(lpBaseOfDll, process = aProcess)
+            aProcess._ModuleContainer__add_module(aModule)
+        return aModule
+
+    def get_file_handle(self):
+        """
+        @rtype:  None or L{FileHandle}
+        @return: File handle to the recently unloaded DLL.
+            Returns C{None} if the handle is not available.
+        """
+        hFile = self.get_module().hFile
+        if hFile in (0, win32.NULL, win32.INVALID_HANDLE_VALUE):
+            hFile = None
+        return hFile
 
     def get_filename(self):
         """
@@ -882,9 +1050,9 @@ class OutputDebugStringEvent (Event):
             It may be ANSI or Unicode and may end with a null character.
         """
         return self.get_process().peek_string(
-                                    self.raw.u.DebugString.lpDebugStringData,
-                                    bool( self.raw.u.DebugString.fUnicode ),
-                                    self.raw.u.DebugString.nDebugStringLength)
+                                self.raw.u.DebugString.lpDebugStringData,
+                                bool( self.raw.u.DebugString.fUnicode ),
+                                self.raw.u.DebugString.nDebugStringLength)
 
 #==============================================================================
 
@@ -1012,7 +1180,8 @@ class EventHandler (object):
        Currently we have partial support for C++ exceptions thrown by Microsoft
        compilers.
 
-       Also see: U{RaiseException()<http://msdn.microsoft.com/en-us/library/ms680552(VS.85).aspx>}
+       Also see: U{RaiseException()
+       <http://msdn.microsoft.com/en-us/library/ms680552(VS.85).aspx>}
 
      - I{create_thread}
 
@@ -1075,14 +1244,16 @@ class EventHandler (object):
      - I{privileged_instruction}
      - I{single_step}
      - I{stack_overflow}
+     - I{wow64_breakpoint}
 
 
 
     @type apiHooks: dict( str S{->} tuple( str, int ) )
     @cvar apiHooks:
-        Dictionary that maps module names to tuples of ( procedure name, parameter count ).
+        Dictionary that maps module names to
+        tuples of ( procedure name, parameter count ).
 
-        All procedures listed here will be hooked for calls from the debuguee.
+        All procedures listed here will be hooked for calls from the debugee.
         When this happens, the corresponding event handler is notified both
         when the procedure is entered and when it's left by the debugee.
 
@@ -1137,6 +1308,11 @@ class EventHandler (object):
     apiHooks = {}
 
     def __init__(self):
+        # XXX HACK
+        # This will be removed when hooks are supported in AMD64.
+        if self.apiHooks and win32.CONTEXT.arch != 'i386':
+            raise NotImplementedError, "Hooks are not yet implemented in 64 bits"
+
         # Convert the tuples into instances of the ApiHook class.
         # A new dictionary must be instanced, otherwise we could also be
         #  affecting all other instances of the EventHandler.
@@ -1154,11 +1330,12 @@ class EventHandler (object):
             fileName = event.get_module().get_filename()
             if fileName:
                 lib_name = PathOperations.pathname_to_filename(fileName).lower()
+                debug    = event.debug
+                pid      = event.get_pid()
                 for hook_lib, hook_api_list in self.__apiHooks.iteritems():
                     if hook_lib == lib_name:
                         for hook_api_stub in hook_api_list:
-                            hook_api_stub.hook(event.debug, event.get_pid(),
-                                                                      lib_name)
+                            hook_api_stub.hook(debug, pid, lib_name)
 
     def __call__(self, event):
         """
@@ -1243,14 +1420,17 @@ class EventDispatcher (object):
             However you'll probably find it more convenient to use an instance
             of a subclass of L{EventHandler} here.
         """
-        if eventHandler is not None:
-            bCallable = hasattr(eventHandler, '__call__')
-            try:
-                bCallable = bCallable or callable(eventHandler)
-            except NameError:
-                pass    # "callable" doesn't exist in Python 3.x
-            if not bCallable:
-                raise TypeError, "Invalid event handler"
+        if eventHandler is not None and not callable(eventHandler):
+            raise TypeError, "Event handler must be a callable object"
+        try:
+            if issubclass(eventHandler):
+                classname = str(eventHandler)
+                msg  = "Event handler must be an instance of class %s"
+                msg += "rather than the %s class itself. Missing brackets?"
+                msg  = msg % (classname, classname)
+                raise TypeError, msg
+        except TypeError:
+            pass
         self.__eventHandler = eventHandler
 
     def dispatch(self, event):
@@ -1275,7 +1455,7 @@ class EventDispatcher (object):
         bCallHandler = True
         pre_handler  = None
         post_handler = None
-        eventCode    = event.get_code()
+        eventCode    = event.get_event_code()
 
         # Get the pre and post notification methods for exceptions.
         # If not found, the following steps take care of that.
